#!/usr/bin/perl
#
# Top-level perl script for CSPP_NUCAPS-IASI package.
# Uses only Perl modules that are part of the v5.006 standard install.
#
# The NOAA/STAR perl scripts have dependencies which I shall have to
# make sure are packaged and accessible.
#
# A set of user-defined functions are appended to this main script;
# they are listed in alphabetical order.
#
# Pre-requisites:
# Distribution from NOAA/STAR in directory 
#   $CSPP_NUCAPS_IASI_HOME/$CSPP_NUCAPS_IASI_REVNAME
#
#
# START-HELP (N.B. This tag is used in printing help, do not touch!)
#
#  Options:
#  --version             show program's version number and exit
#  ?, -h, --help         show this help message and exit
#  -n, --nothing         make the PCF files, but do not execute.
#
#  Mandatory Arguments:
#    At a minimum these arguments must be specified [i.e. no default].
#
#    -i INPUT_FILES, --input_files=INPUT_FILES
#                        Path to input files. May be directory name, or
#                        directory/filemask (use * and ?, not regexp).
#                        Wildcarded filemasks MUST be "quoted" (or OS expands).
#                        e.g. -i "/dataDir/*M01_20180418*" could work, but
#                        depends on what else is in the directory.
#                        (BUT -i /dataDir  really is the simplest way).
#    -s MISSION, --sat=MISSION
#                        The mission satellite data for NUCAPS algorithm.
#                        Possible values are: 'metopA' and 'metopB'.
#
#  Extra Options:
#    These options may be used to customize behaviour of this program.
#
#    -w WORK_DIR, --work=WORK_DIR
#                        The directory in which all activity will occur,
#                        this is where output files will be put.
#                        [default: current directory].
#    -m MDR, --mdrstart=MDR
#                        MDR to start from in slicing the DB overpass into
#                        IASI L1C granules.  Possible values are '0' (first),
#                        some other integer, 'S' meaning prefer the start of
#                        the overpass (i.e.the same as '0'), 'E' meaning prefer
#                        the end of the overpass, or 'M' meaning preferentially
#                        choose the middle of overpass.
#                        [default: 'M'].
#    -d DYNAMIC_ANC, --dynanc=DYNAMIC_ANC
#                        Path to root of dynamic ancillary data tree;
#                        beneath here ancillary data are organized as
#                        per the SSEC remote ancillary server for CSPP,
#                        https://jpssdb.ssec.wisc.edu/cspp_v_2_0/ancillary/,
#                        i.e. %Y_%m_%d_%j where %Y is four-digit year, %m is
#                        two-digit month, %d is two-digit day of month and
#                        %j is three-digit day of year.
#                        [CSPP_NUCAPS_IASI_HOME/data/dynamic_ancillary]
#    -p PROCESSORS, --processors=PROCESSORS
#                        Number of cpus to use for NUCAPS-IASI processing.
#                        [default: --processors=1].
#    -v, --verbose       Each occurrence increases verbosity 1 level from
#                        [default: ERROR only] -v=WARNING -vv=INFO -vvv=DEBUG
#    -c, --clean         Deletes temporary working directory under WORK_DIR
#                        at process completion.
#                        [default: --clean; --noclean is the contrary option].
#    -q, --quarantine    Processing machine is not connected to the web for the
#                        purpose of fetching required dynamic ancillary data.
#                        [default: --noquarantine].
#    -x, --xfiles        Use non-DB filenames for AMSUA/MHS L1B data sources.
#                        [default: Direct Broadcast filenames].
#
#
# END-HELP (N.B. This tag is used in printing help, do not touch!)
#
# The following support has been dropped until I know what data sources are
# allowed and how they can be mixed & matched.
#
#    -a, --aggregate     Aggregates granules output into swaths of up
#                        to 30 minutes as a post-processing step.
#                        [default: --noagg].
#    -r GRIB, --reproc=GRIB
#                        Reprocessing mode using GFS data in grib version GRIB,
#                        where GRIB is 1 or 2. Activating this option also sets
#                        the GFS local repo to DYNAMIC_ANC/%Y/%m (see above)
#                        AND forces --quarantine (see below).
#                        [default: --reproc=0 meaning reprocessing mode OFF].
#
# NOTES-TO-SELF:
# -------------
#   $!=N to change exit code on die() to N; default is 255
#
#   --clean removes the temporary directory tree. If the user selects --noclean
#   the temporary directory is NOT removed and all NUCAPS product files are 
#   available in the temporary directory structure. This option might be useful
#   for debugging and is necessary if the user wants access to the product files
#   generated prior to the output netCDF files (that are copied up to WORK_DIR).
#   It is then user responsibility to dispense with unwanted directories/files.
#
# A design decision (that might get revoked) is to require input data to be in
# a single directory (even if user is using a filemask, you presently cannot
# escape this - see use of $options{input_dir} in function validate_options().
#
# There is a top-level script provided by NOAA/STAR called run_ISRP.pl which
# accepts on its commandline a configuration file e.g. ISRP_CONFIG, and a date
# e.g. 20180228.
#
# So most of the work has been done by NOAA/STAR.  But, to make the system
# fit the CSPP paradigm, I still have some work ahead of me.
#
# I have included a utility from Eumetsat called kai to slice up the DB IASI
# files.  Where shall I put it?  I think I will just put the executable in the
# bin/ directory alongside ...
# h5diff h5dump h5import h5ls h5repack h5stat nagg ncdump ncgen wgrib wgrib2
# ... but I need to put a build directory somewhere, let's say here:
# /data/jimd/CSPP_Projects/CSPP_NUCAPS-IASI/kai/
#
#
# DESIGN PHILOSOPHY (is there one?)
# ---------------------------------
#
# My CSPP front end script must look in the specified directory for files to
# process, go fetch dynamic ancillary data from our server and put them where
# they need to be, create a temporary directory to work in (outside of the
# distribution tree), run run_ISRP.pl,move products out to some specified (or
# default) directory and finally disappear the temporary working directory
# (or optionally leave it for debugging).  NOTE: I have decided to not run
# run_ISRP.pl but separately run the Perl scripts that it comprises.
#
# I will probably have to write ISRP_CONFIG from a template on the fly, but
# otherwise not change too much else.  I might need to hack the NOAA/STAR
# scripts on the fly too if I hit a problem with long absolute paths.  It
# looks like the NOAA/STAR system chdir's into the Granule-??? directories
# that it creates, so I'll bet that I can make file references local.  The
# problem will need to be solved in some dynamic way because I think some
# paths in ISRP_CONFIG will need to be absolute during soem part of the
# processing, but could be relative in another (to avoid overflowing the
# command line buffer of some executables with long absolute path names).
#
# I need to check how parallelism is being implemented by NOAA/STAR and my
# preference is to take control of how many cores are employed.
#
# For CSPP I have to make sure that no files are written to in the distribution
# tree (so that user B can run user A’s installation; and it also make a big
# difference to cluster processing).  Generally this just means copying more
# out of the distribution tree and into the temporary work directory (and/or
# creating a directory structure in the temporary working directory to allow
# the system to work).  One of the problems that has caught me off-guard in the
# past is files in the distribution tree that only need to be read, but the
# programmer has them opened as ‘rw’.
#
# James E Davies SSEC/UW-Madison  Apr/May 2018 for version 1.0
#


# Pragmata
# ````````
use 5.006;
use strict;
use warnings;


# Load Perl modules - standard modules only
# `````````````````````````````````````````
use POSIX;
use Storable;
use File::Spec;
use File::Copy;
use Data::Dumper;
use Getopt::Long;
use File::Basename;
use Cwd qw(abs_path);
use Encode qw(decode);
use Time::gmtime qw(:FIELDS);
use List::Util qw(sum min max);
use File::Path qw(mkpath rmtree);
use Time::Local qw(timegm_nocheck);
use File::Temp qw(tempfile tempdir);
use Scalar::Util qw(looks_like_number);
use I18N::Langinfo qw(langinfo CODESET);


# Global variables are all UPPERCASE throughout
# `````````````````````````````````````````````
our $AGG_WINDOW_MINUTES = 30; # 30-minute aggregation window for npp
our $SLEEP_GFS_MINUTES = 0;   # 0 means disable wait for preferred GFS data
our $MAIN = basename($0);
defined $ENV{CSPP_NUCAPS_IASI_HOME} or die "$MAIN\:: CSPP_NUCAPS_IASI_HOME not set.\nAbort";
defined $ENV{CSPP_NUCAPS_IASI_REVNAME} or die "$MAIN\:: CSPP_NUCAPS_IASI_REVNAME not set.\nAbort";
defined $ENV{CSPP_NUCAPS_IASI_REVDATE} or die "$MAIN\:: CSPP_NUCAPS_IASI_REVDATE not set.\nAbort";
our $DAP = File::Spec -> catfile($ENV{CSPP_NUCAPS_IASI_HOME}, $ENV{CSPP_NUCAPS_IASI_REVNAME});
our $HOME = File::Spec -> catfile($ENV{CSPP_NUCAPS_IASI_HOME}, 'bin');
(-d $HOME) or die "$MAIN\:: CSPP_NUCAPS-IASI subdirectory $HOME not found.\nAbort";


# $WGRIB points to supplied static-compiled wgrib (v1.8.1.2a)
# ```````````````````````````````````````````````````````````
our $WGRIB = File::Spec -> catfile($ENV{'CSPP_NUCAPS_IASI_HOME'}, 'bin', 'wgrib');
(-e $WGRIB) or die "$MAIN\:: h5repack $WGRIB not found.\nAbort";
(-x $WGRIB) or die "$MAIN\:: h5repack $WGRIB not executable.\nAbort";


# $WGRIB2 points to supplied static-compiled wgrib2 (v0.1.9.5.1)
# ``````````````````````````````````````````````````````````````
our $WGRIB2 = File::Spec -> catfile($ENV{'CSPP_NUCAPS_IASI_HOME'}, 'bin', 'wgrib2');
(-e $WGRIB2) or die "$MAIN\:: h5repack $WGRIB2 not found.\nAbort";
(-x $WGRIB2) or die "$MAIN\:: h5repack $WGRIB2 not executable.\nAbort";


# $KAI points to supplied dymamic-compiled kai
# ````````````````````````````````````````````
# Compiled from kai-1.11.tar.gz supplied by Simon.Elliott@eumetsat.int
our $KAI = File::Spec -> catfile($ENV{'CSPP_NUCAPS_IASI_HOME'}, 'bin', 'kai');
(-e $KAI) or die "$MAIN\:: kai $KAI not found.\nAbort";
(-x $KAI) or die "$MAIN\:: kai $KAI not executable.\nAbort";


# Get the command line options, display help or revision if asked
# ```````````````````````````````````````````````````````````````
my $revision = "CSPP_NUCAPS-IASI v1.0 $ENV{CSPP_NUCAPS_IASI_REVDATE}";
my $pwd = getcwd();
my %options = get_commandline($revision, $pwd);
($options{verbose} >= 2) and pretty_dump('%options', %options);


# uuidgen is a system requirement
# ```````````````````````````````
our $UUIDGEN = `which uuidgen`; chomp($UUIDGEN);
(index($UUIDGEN, 'uuidgen') < 0) and
  die "$MAIN\:: Required utility uuidgen not found\nAbort"; 


# At present we are saying that lftp is a system requirement
# ``````````````````````````````````````````````````````````
our $LFTP = 'lftp';
($options{quarantine}) or do {
  system("$LFTP --version 1>/dev/null 2>/dev/null") == 0
    or die "$MAIN\:: Required utility lftp not found\nAbort"; };


# $NCDUMP points to supplied static-compiled ncdump (netcdf library version 4.3.0)
# ````````````````````````````````````````````````````````````````````````````````
our $NCDUMP = File::Spec -> catfile($ENV{'CSPP_NUCAPS_IASI_HOME'}, 'bin', 'ncdump');
($options{aggregate}) and do {
  (-e $NCDUMP) or die "$MAIN\:: $NCDUMP not found.\nAbort";
  (-x $NCDUMP) or die "$MAIN\:: $NCDUMP not executable.\nAbort"; };


# $NCGEN points to supplied static-compiled ncgen (netcdf library version 4.3.0)
# ``````````````````````````````````````````````````````````````````````````````
our $NCGEN = File::Spec -> catfile($ENV{'CSPP_NUCAPS_IASI_HOME'}, 'bin', 'ncgen');
($options{aggregate}) and do {
  (-e $NCGEN) or die "$MAIN\:: $NCGEN not found.\nAbort";
  (-x $NCGEN) or die "$MAIN\:: $NCGEN not executable.\nAbort"; };


# These are used for fileglobbing when INPUTFILES is a directory name
# ```````````````````````````````````````````````````````````````````
# Change to suit your site; for more control edit function validate_options(),
# or provide "<dirname>/<my_filemask>" on the command line as INPUTFILES; the
# "quotes" are important to prevent the OS expanding wildcard tokens. This is
# also where you can encode for different DB and DAAC naming conventions.
my %filemask = (metopA => '{{amsua,mhs}l1b_M02_*.l1b,IASI_xxx_1C_M02_*}',
                metopB => '{{amsua,mhs}l1b_M01_*.l1b,IASI_xxx_1C_M01_*}');
($options{xfiles}) and do { # i.e. non-DB filenaming conventions
  %filemask = (metopA => '{*.{AMAX,MHSX}.M2.*,IASI_xxx_1C_M02_*}',
               metopB => '{*.{AMAX,MHSX}.M1.*,IASI_xxx_1C_M01_*}'); };
# Debug information
($options{verbose} == 3) and pretty_dump('%filemask', %filemask);


# Parse and validate command line %options & make temporary working directory
# ```````````````````````````````````````````````````````````````````````````
my $sat = $options{sat};
%options = validate_options($filemask{$sat}, %options);
($options{verbose} == 3) and pretty_dump('%options', %options);
my $tmpdir = tempdir('TMP_XXXXXX', DIR => $options{work},
  CLEANUP => $options{clean});
chdir($tmpdir) or die "$MAIN\:: Cannot chdir to temp dir $tmpdir\nAbort";
($options{verbose} >= 2) and print "chdir to working directory $tmpdir\n";


# Make directories to hold input data that have been converted as required
# ````````````````````````````````````````````````````````````````````````
mkpath('input') or die "$MAIN\:: WORK_DIR/input cannot be made.\nAbort";
foreach ( qw|L1B/AMSUA L1B/MHS L1C/IASI| ) {
  mkpath("input/$_") or die "$MAIN\:: WORK_DIR/input/$_ cannot be made.\nAbort";
}


# Convert the specified @options{input_files} to required filenames and formats
# `````````````````````````````````````````````````````````````````````````````
foreach (@{$options{input_files}}) {
  my $fullfile = File::Spec -> catfile($options{input_dir}, $_);
  # AMSUA DB - make sure big endian and NOAA name
  (index($_, 'amsua') < 0) or do {
    (length($_)==36) or next;
    my $outfile = make_L1B_bigEndian('amsua', $fullfile);
    my $inplace = 'NSS'.substr($outfile, 3);
    $inplace =~ s/X.MB.D/X.M1.D/g; $inplace =~ s/X.MA.D/X.M2.D/g;
    move($outfile, "input/L1B/AMSUA/$inplace") or warn "$MAIN\:: Could not move $outfile\n"; };
  # AMSUA NOAA - symlink
  (index($_, 'AMAX') < 0) or do {
    (length($_)==42) or next;
    symlink($fullfile, "input/L1B/AMSUA/$_") or warn "$MAIN\:: Could not symlink $_\n"; };
  # MHS DB - make sure big endian and NOAA name
  (index($_, 'mhs') < 0) or do {
    (length($_)==34) or next;
    my $outfile = make_L1B_bigEndian('mhs', $fullfile);
    my $inplace = 'NSS'.substr($outfile, 3);
    $inplace =~ s/X.MB.D/X.M1.D/g; $inplace =~ s/X.MA.D/X.M2.D/g;
    move($outfile, "input/L1B/MHS/$inplace") or  warn "$MAIN\:: Could not move $outfile\n"; };
  # MHS NOAA - symlink
  (index($_, 'MHSX') < 0) or do {
    (length($_)==42) or next;
    symlink($fullfile, "input/L1B/MHS/$_") or warn "$MAIN\:: Could not symlink $_\n"; };
  # IASI
  (index($_, 'IASI_xxx_1C') < 0) or do {
    if ($options{xfiles}) {
      # NOAA - symlink
      symlink($fullfile, "input/L1C/IASI/$_") or warn "$MAIN\:: Could not symlink $_\n";
    } else {
      # DB - slice up with kai then move
      my $kai_skip = compute_kai_skip($options{mdrstart}, $fullfile);
      my $cmd = "$KAI --pdu XXX --skip $kai_skip --in $fullfile 2>/dev/null";
      my $dribble = `$cmd`;
      my ($n, $undersize) = (0, 1);
      foreach (split("\n",$dribble)) {
        (index($_, 'Setting MPHR TOTAL_MDR') < 0) or do {
          my $mdr = trim_string((split(' to ', $_))[1]);
          (grep {$_ eq $mdr} qw/22 23/) and $undersize = 0; };
        (index($_, 'Setting MPHR PRODUCT_NAME') < 0) or do {
          my $granule = trim_string((split(' to ', $_))[1]);
          my $infile = sprintf("XXX.%05d.pdu", $n++);
          ($undersize) and do { unlink($infile); next };
          $undersize = 1;
          my $inplace = "input/L1C/IASI/$granule";
          move($infile, $inplace) or warn "$MAIN\:: Could not move $infile\n";
        };
      }
    }
  };
}


# What dates are covered by the files to process, just use IASI files
# ```````````````````````````````````````````````````````````````````
my $globber = File::Spec -> catfile('input/L1C/IASI', 'IASI_xxx_1C*');
my %uniqdate = (); my %gmthash = ();
foreach (sort(glob($globber))) {
  my $X = substr(basename($_), 16, 8);
  my $Y = substr(basename($_), 24, 6);
  my ($YYYY, $MM, $DD) = (substr($X, 0, 4), substr($X, 4, 2), substr($X, 6, 2));
  my ($hh, $mm, $ss) = (substr($Y, 0, 2), substr($Y, 2, 2), substr($Y, 4, 2));
  my $Z = timegm_nocheck($ss, $mm, $hh, $DD, $MM-1, $YYYY);
  $uniqdate{$X} = 'unique'; # YYYYMMDD, use hash key to force unique list
  $gmthash{$Z} = 'unique'; # use hash key to force unique list
}
my @isrp_date = sort keys %uniqdate; # this is needed for ISRP.pl command line
my @gmt = sort keys %gmthash; # this is needed for fetching/staging GFS data


# Define dynamic data remote & local locations and file naming conventions
# ````````````````````````````````````````````````````````````````````````
# The root of the local directory for dynamic ancillary data is either defined
# on the command line or is assigned a default value. I am appending the agreed
# subdirectory naming convention to this root and copying it across to a hash
# that contains much of what I need to know to fetch and store ancillary data.
# See https://jpssdb.ssec.wisc.edu/cspp_v_2_0/ancillary/ for directory structure.
# I have added support for GRIB1 files and an alternative repo directory for
# reprocessing.
my $usual_repo = File::Spec -> catfile($options{dynanc}, '%Y_%m_%d_%j');
my $repro_repo = File::Spec -> catfile($options{dynanc}, '%Y/%m');
my $local_repo = $usual_repo;
($options{reproc}) and $local_repo = $repro_repo;
my $grib1_template = 'gblav.PGrbF%f.%y%m%d.%Hz';
my $grib2_template = 'gfs.t%H.%y%m%d.pgrbf%f';
my $gfs_template = $grib2_template;
($options{reproc} == 1) and $gfs_template = $grib1_template;
my %dynanc_loc = (GFSPROTO => 'https://',
                  GFSSITE => 'jpssdb.ssec.wisc.edu',
                  GFSREMOTE => 'cspp_v_2_0/ancillary/%Y_%m_%d_%j',
                  GFSTEMPLATE => $gfs_template,
                  GFSLOCAL => $local_repo);


# Fetch GFS and wrangle into place for all dates to be processed
# ``````````````````````````````````````````````````````````````
populate_dynanc_gfs(\%dynanc_loc, \%options, @gmt);


# Make the CPU_STATUS file in $tmpdir to indicate how many parallel streams
# `````````````````````````````````````````````````````````````````````````
make_CPU_STATUS(%options);


# Make the ISRP_CONFIG file (OK for multiple days) and link/copy other support
# ````````````````````````````````````````````````````````````````````````````
make_ISRP_CONFIG(%options);
mkpath("time_files") or die "$MAIN\:: WORK_DIR/time_files cannot be made.\nAbort";
foreach (glob("$DAP/sys_files/time_files/*")) {
  copy($_, 'time_files');
}


# Make symlinks in $tmpdir to support files
# `````````````````````````````````````````
# N.B. $PLATFORM_SYS and sys_files have been symlinked in the temporary
# working directory $tmpdir because long filepaths to support files causes
# some (at least one) DAP executable to overflow its command line buffer.
symlink("$DAP/sys_files", 'sys_files') or die "$MAIN\:: Could not symlink sys_files\n";
($options{sat} eq 'metopA') and do { symlink("$DAP/metopA_sys_files", 'metopA_sys_files')
  or die "$MAIN\:: Could not symlink metopA_sys_files\n"; };
($options{sat} eq 'metopB') and do { symlink("$DAP/metopB_sys_files", 'metopB_sys_files')
  or die "$MAIN\:: Could not symlink metopB_sys_files\n"; };


# Main loop on date; this will almost always be just one day!
# ```````````````````````````````````````````````````````````
my $now = time();
(($options{verbose} >= 1) and $options{nothing}) and do {
  print "Option --nothing selected. No package processing scripts will be run,\n";
  print "just the preparations to run. Keep the temporary working directory by\n";
  print "using command line switch --noclean\n"; };
my $go = !$options{nothing};
foreach (@isrp_date) {
  my ($YYYY, $MM, $DD) = (substr($_, 0, 4), substr($_, 4, 2), substr($_, 6, 2));
  my $time = timegm_nocheck(0, 0, 0, $DD, $MM-1, $YYYY);
  gmtime($time);
  my $ready = sprintf("D%02d%03d", $tm_year % 100, $tm_yday + 1);
  my $agedays = int(($now - $time)/60/60/24 + 2);
  # Error reporting left to the ISRP logging system, so no "or warn/die" here
  ($go & ($options{verbose} >= 1)) and print "Run_Check_L1C.pl\n";
  $go and system("perl $DAP/scripts/perl/Run_Check_L1C.pl $tmpdir/ISRP_CONFIG");
  make_CPU_STATUS(%options);
  ($go & ($options{verbose} >= 1)) and print "Process_Status.pl\n";
  $go and system("perl $DAP/scripts/perl/Process_Status.pl $tmpdir/ISRP_CONFIG $agedays 2 10");
  make_CPU_STATUS(%options);
  ($go & ($options{verbose} >= 1)) and print "Run_IASI_L1C_Subsetter.pl\n";
  $go and system("perl $DAP/scripts/perl/Run_IASI_L1C_Subsetter.pl $tmpdir/ISRP_CONFIG");
  make_CPU_STATUS(%options);
  # The touch needs to be done close in time to when files are needed
  foreach (glob("input/L1B/*/*.$ready.*")) { system("touch $_.Ready"); }
  ($go & ($options{verbose} >= 1)) and print "Run_Check_AMSUA_L1B.pl\n";
  $go and system("perl $DAP/scripts/perl/Run_Check_AMSUA_L1B.pl $tmpdir/ISRP_CONFIG $agedays");
  make_CPU_STATUS(%options);
  ($go & ($options{verbose} >= 1)) and print "Run_Check_MHS_L1B.pl\n";
  $go and system("perl $DAP/scripts/perl/Run_Check_MHS_L1B.pl $tmpdir/ISRP_CONFIG $agedays");
  make_CPU_STATUS(%options);
  ($go & ($options{verbose} >= 1)) and print "Run_AMSUA_L1B.pl\n";
  $go and system("perl $DAP/scripts/perl/Run_AMSUA_L1B.pl $tmpdir/ISRP_CONFIG");
  make_CPU_STATUS(%options);
  ($go & ($options{verbose} >= 1)) and print "Run_MHS_L1B.pl\n";
  $go and system("perl $DAP/scripts/perl/Run_MHS_L1B.pl $tmpdir/ISRP_CONFIG");
  make_CPU_STATUS(%options);
  ($go & ($options{verbose} >= 1)) and print "Run_IASI_Level2_Processing.pl\n";
  $go and system("perl $DAP/scripts/perl/Run_IASI_Level2_Processing.pl $tmpdir/ISRP_CONFIG");
  # Move files out to work directory which is up one level from $tmpdir
  my @logs_to_move = glob('logs/*/*/*/*');
  ($go & ($options{verbose} >= 1)) and print "Moving log files to WORK_DIR\n";
  foreach (@logs_to_move) {
    move($_, '..') or do {
      ($options{verbose} >= 1) and warn "WARNING: Move failed: $!"; };
  }
  my @product_to_move = glob("retrieval/$YYYY/$MM/$DD/IASI_DATA/*/level2/IASI*.nc");
  ($go & ($options{verbose} >= 1)) and print "Moving product files to WORK_DIR\n";
  foreach (@product_to_move) {
    # check main_iasi_ccr_to_netcdf.log / main_iasi_l2_to_netcdf.ret.log to validate
    my $log = 'nolog';
    (index(basename($_), 'IASI_L02') == 0) and $log = 'main_iasi_l2_to_netcdf.ret.log';
    (index(basename($_), 'IASI_CCR') == 0) and $log = 'main_iasi_ccr_to_netcdf.log';
    $log = File::Spec -> catfile(dirname($_), $log);
    -e $log and do {
      open (my $fh,"<$log"); my @lines=<$fh>; close $fh;
      my @find = grep (/FATAL/, @lines);
      ($#find >= 0) or do {
        move($_, '..') or do {
          ($options{verbose} >= 1) and warn "WARNING: Move failed: $!"; };
      };
    };
  }
  # Do not need to remove input/L1B/*/*.Ready files because they go stale fast
}


# All done, chdir to starting directory so that $tmpdir can be removed
# ````````````````````````````````````````````````````````````````````
if (not $options{clean}) {
  print "--noclean set, reminder: tmpdir is ...\n$tmpdir\n";
}
chdir($pwd) or die "$MAIN\:: Cannot chdir back to $pwd\nAbort";
exit(0);


###############################################################################
#                                                                             #
#         User-defined functions - they are listed alphabetically             #
#                                                                             #
#   NOTE: My editor can collapse code between matching {} braces, this        #
#   explains the strange formatting - I can collapse each function to see     #
#   just the function name and an example of how to call it; and expand       #
#   to see the more detailed documentation and to edit.      JED Nov 2013     #
#                                                                             #
###############################################################################

# ------------------
  sub aggregate_data
# ==================
# @data = aggregate_data($varname, @granule_filenames);
  {
    # @data is a vector of strings that comprise the data from
    # all @granule_filenames for specific $varname and in a form
    # that can be appended to the .cdl under creation.
    # NOTE: I think it is cheating, but by changing the Pressure and
    # Effective_Pressure dimensions to 1D [in aggregate_varatts()] it appears
    # to be unnecessary to limit the pressure data to one profile as
    # ncgen discards all but the first profile - which is basically what I
    # want without all the work of hacking the data stream.
    my $varname = shift;
    my @files = sort(@_);
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my @data = ();
    foreach my $fullname (@files) {
      my $key = (split('\.', basename($fullname)))[0];
      my $cmd = "$NCDUMP -v $varname $fullname |";
      open(my $fh, $cmd) or die "$name\: Could not run command $cmd\nAbort";
      while (<$fh>) { chomp; ( /data:/ ) and last; }
      my @buffer = ();
      while (<$fh>) { chomp; ( /}/ ) and last; push(@buffer, $_); }
      close($fh);
      foreach (@buffer) {
        my $value = '';
        my @parts = split(/=/, $_);
        ($#parts == -1) and next;
        ($#parts ==  0 and /=/) and next;
        ($#parts ==  0) and push(@data, (split(/;/, $parts[0]))[0]);
        ($#parts ==  1) and push(@data, (split(/;/, $parts[1]))[0]);
      }
      ($fullname eq $files[-1]) and $data[-1] = $data[-1]." ;";
      ($fullname ne $files[-1]) and $data[-1] = $data[-1].",";
    }
    return @data;
  }

# ----------------------
  sub aggregate_fileroot
# ======================
# $fileroot = aggregate_fileroot(@granule_filenames);
  {
    # The returned fileroot includes neither the file path nor the extension
    # It is up to the user to check that @granule_filenames has at least
    # one element. If the filenames do not break on '_' as expected, this
    # routine will bail out with an error message.
    my @files = sort(@_);
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my $first = basename($files[0]);
    my $last = basename($files[$#files]);
    my @first = split(/\_/, $first);
    my @last = split(/\_/, $last);
    ($#first == 5) or die "$name\: first file $first invalid\nAbort";
    ($#last == 5) or die "$name\: last file $last invalid\nAbort";
    gmtime();
    my $creation = strftime("c%Y%m%d%H%M%S0", $tm_sec, $tm_min, $tm_hour,
      $tm_mday, $tm_mon, $tm_year);
    my $fileroot = join ( '_', @first[0..3], $last[4], $creation);
    return $fileroot;
  }

# ----------------------
  sub aggregate_globatts
# ======================
# %globatt = aggregate_globatts($attname, $opcode, \%headers, %globatt);
  {
    # %globatt is updated with a key=$attname and a value=full attrbute line
    # Supported opcodes are 'start', 'finish', 'list', 'polygon'
    # NOTE: There is nothing to be gained in call this routine if the number
    # of %input_cls is just ONE. In fact there would have been no purpose in
    # even populating %input_cls  SO  this routine ASSUMES there are at least
    # TWO files represented in %input_cdls.
    my ($attname, $opcode) = (shift, shift);
    my %headers = %{(shift)};
    my (%globatt) = @_;
    my @values = ();
    foreach my $key (sort keys %headers) {
      my @buffer = @{$headers{$key}};
      my $i = 0;
      foreach (@buffer) { ( /global attributes:/ ) and last; $i++; }
      my $j = $i;
      foreach (@buffer[$i..$#buffer]) { ( /$attname/ ) and last; $j++; }
      push(@values, $buffer[$j]);
    }
    ($opcode eq 'start') and $globatt{$attname} = $values[0];
    ($opcode eq 'finish') and $globatt{$attname} = $values[-1];
    ($opcode eq 'list') and do {
      my @part = split(/=/, $values[0]);
      my $item = trim_string((split(';', $part[1]))[0]);
      looks_like_number($item) or $item = (split(/\"/, $item))[1];
      $globatt{$attname} = sprintf("%s  = \"LIST((%s", $part[0], $item);
      foreach (@values[1..$#values]) {
        $item = trim_string((split(';', (split(/=/, $_))[1]))[0]);
        looks_like_number($item) or $item = (split(/\"/, $item))[1];
        $globatt{$attname} = sprintf("%s, %s", $globatt{$attname}, $item);
      }
      $globatt{$attname} = sprintf("%s))\" ;", $globatt{$attname});
    };
    ($opcode eq 'polygon') and do {
      my %polygon = ();
      for (my $i=0; $i<=$#values; $i++) {
        my $q1 = (split('\(\(', $values[$i]))[1];
        my $q2 = (split('\)\)', $q1))[0];
        my @q3 = split(',', $q2);
        @{$polygon{$i}} = @q3;
      }
      my $n = scalar(keys %polygon) - 1; # largest key in %polygon
      my $newvalue = "\t\t".':geospatial_bounds = "POLYGON((';
      $newvalue = $newvalue.join(',', @{$polygon{0}}[0..2]);
      foreach my $key (sort keys %polygon) {
        $newvalue = join(',', $newvalue, @{$polygon{$key}}[3..4]);
      }
      $newvalue = join(',', $newvalue, @{$polygon{$n}}[5..6]);
      foreach my $key (reverse sort keys %polygon) {
        $newvalue = join(',', $newvalue, @{$polygon{$key}}[7..8]);
      }
      $newvalue = $newvalue.'))" ;';
      $globatt{$attname} = $newvalue;
    };
    return %globatt;
  }

# --------------------
  sub aggregate_header
# ====================
# @header = aggregate_header(\%headers, \%newDim, \%varatt, %globatt);
  {
    # @header is the aggregated header based on the individual granule
    # headers in %headers, the new dimensions information in %newDim, the
    # updates to variable attributes in %varatt and the global attributes
    # modified where appropriate to represent the entire aggregated file
    # in %globatt.
    my %headers = %{(shift)};
    my %newDim = %{(shift)};
    my %varatt = %{(shift)};
    my (%globatt) = @_;
    my @header = ();
    my $filekey =  (sort(keys %headers))[0];
    my @buffer = @{$headers{$filekey}};
    my $section = 0;
    foreach (@buffer) {
      /$filekey/ and next; # because name has been replaced with $ccr_aggregate
      /Number_of_CrIS_FORs = 120 ;/ and next;
      /Scan_Line/ and next;
      /\tint CrIS_FORs/ and next;
      /\t\tCrIS_FORs:/ and next;
      /dimensions:/ and $section = 1;
      /variables:/ and $section = 2;
      /global attributes:/ and $section = 3;
      my $value = $_;
      ($section == 1) and do {
        /dimensions:/ and do { foreach (@{$newDim{dim}}) { $value = $value."\n".$_; } }; };
      ($section == 2) and do {
        /variables:/ and do { foreach (@{$newDim{var}}) { $value = $value."\n".$_; } };
        my $verify = 'N/A';
        if ( /=/ ) {
          $verify = trim_string((split(/=/, $_))[0]);
        } else {
          ( /\t/ ) and do {
            $verify = trim_string((split(/ /, $_))[1]);
            $verify =~ s/\(/\\\(/g;  # Need to escape the brackets otherwise the
            $verify =~ s/\)/\\\)/g;  #   search/replace fails in aggregate_header()
          };
        }
        foreach my $key (sort keys %varatt) {
          ($key eq $verify) and do { $value = $varatt{$key}; last; };
        }
      };
      ($section == 3) and do {
        foreach my $key (sort keys %globatt) {
          /$key/ and do { $value = $globatt{$key}; last; };
        }
      };
      push (@header, $value);
    }
    my $value = sprintf("\t\t:aggregated_by = \"%s\" ;", $globatt{aggregate});
    push (@header, $value);
    gmtime();
    my $creation_date = sprintf("%04d-%02d-%02d %02d:%02d:%02d GMT",
      $tm_year+1900, $tm_mon+1, $tm_mday, $tm_hour, $tm_min, $tm_sec);
    $value = sprintf("\t\t:creation_date = \"%s\" ;", $creation_date);
    push (@header, $value);
    return @header;
  }

# --------------------
  sub aggregate_newDim
# ====================
#  %newDim = aggregate_newDim(@granule_filenames);
  {
    # %newDim is a hash containg the extra information needed to build
    # an extra dimension in the aggregated file, that being simply the
    # number of granules.  It has three keys, {dim} goes in the dimensions:
    # section of the .cdl file, {var} goes in the variables: section and
    # {dat} in the data: section.
    my @files = @_;
    my $n = $#files + 1;
    my $p = 30;
    my $l = $n * 120 / $p;
    my $m = $l * $p;
    my %info = ();
    my $dimy = 'Number_of_CrIS_scans';
    my $dimx = 'Number_of_CrIS_FOR';
    @{$info{dim}} = ();
    push(@{$info{dim}}, sprintf("\t%s = %d ;", $dimy, $l));
    push(@{$info{dim}}, sprintf("\t%s = %d ;", $dimx, $p));
    @{$info{var}} = ();
    my $var = 'Scan_Number';
    push(@{$info{var}}, "\tint $var($dimy, $dimx) ;");
    push(@{$info{var}}, "\t\t$var:long_name = \"Current scan in aggregate\" ;");
    push(@{$info{var}}, sprintf("\t\t%s:valid_range = %d, %d ;", $var, 1, $l));
    push(@{$info{var}}, sprintf("\t\t%s:_FillValue = -9999 ;", $var));
    my @lin = sort {$a <=> $b} ((1..$l) x $p);
    $info{dat} = sprintf(" %s = 1", $var);
    for (my $j=1; $j<$m; $j++) {
      $info{dat} = sprintf("%s, %d", $info{dat}, $lin[$j]);
    }
    $info{dat} = sprintf("%s ;\n", $info{dat});
    $var = 'FOR_Number';
    push(@{$info{var}}, "\tint $var($dimy, $dimx) ;");
    push(@{$info{var}}, "\t\t$var:long_name = \"Current F.O.R. in aggregate\" ;");
    push(@{$info{var}}, sprintf("\t\t%s:valid_range = %d, %d ;", $var, 1, $p));
    push(@{$info{var}}, sprintf("\t\t%s:_FillValue = -9999 ;", $var));
    my @pix = (1..$p) x $l;
    $info{dat} = sprintf("$info{dat}\n %s = 1", $var);
    for (my $j=1; $j<$m; $j++) {
      $info{dat} = sprintf("%s, %d", $info{dat}, $pix[$j]);
    }
    $info{dat} = sprintf("%s ;", $info{dat});
    return %info;
  }

# ------------------------------
  sub aggregate_nucaps_filegroup
# ==============================
#  aggregate_nucaps_filegroup($destDir, $revision, @files);
  {
    # $destDir is the destination directory for aggregated output files
    # $revision is the revision number of this script for global attributes
    # @files is the full pathname of the CCR or EDR files, the
    # code in this routine works for both sets of files!
    my ($destDir, $revision) = (shift, shift);
    my @files = @_;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my $aggregated_filename = aggregate_fileroot(@files);
    # read headers of all files in filegroup
    my %headers = ();
    foreach my $fullname (@files) {
      my $key = (split('\.', basename($fullname)))[0];
      my $cmd = "$NCDUMP -h $fullname |";
      open(my $fh, $cmd) or die "$name\: Could not run command $cmd\nAbort";
      my @buffer = ();
      while (<$fh>) { chomp; push(@buffer, $_); }
      close($fh);
      @{$headers{$key}} = @buffer[0..($#buffer-1)]; # clip the trailing '}'
    }
    # define the new dimension that need to be inserted in aggregate
    my %newDim = aggregate_newDim(@files); # UNLIMITED dimension
    # variable attributes that need changing are keyed by their names
    my %varatt = ();
    %varatt = aggregate_varatts('coordinates', \%headers, %varatt);
    %varatt = aggregate_varatts('dimensions', \%headers, %varatt);
    %varatt = aggregate_varatts('quality', \%headers, %varatt);
    ### pretty_dump('%varatt', %varatt);
    # variable names for later writing into data section
    my @varnames = aggregate_varnames(%headers);
    ### print "@varnames\n";
    # global attributes that need changing are also keyed by their names
    my %globatt = (); # values are opcodes that get replaced by attributes
    $globatt{starting_and_ending_orbit_node} = 'list' ;
    $globatt{date_created} = 'start' ;
    $globatt{Metadata_Link} = 'list' ;
    $globatt{source} = 'list' ;
    $globatt{start_orbit_number} = 'start' ;
    $globatt{end_orbit_number}   =  'finish' ;
    $globatt{day_night_data_flag} = 'list' ;
    $globatt{ascend_descend_data_flag} = 'list' ;
    $globatt{time_coverage_start} = 'start' ;
    $globatt{time_coverage_end}   =  'finish' ;
    $globatt{geospatial_first_scanline_first_fov_lat} = 'start' ;
    $globatt{geospatial_first_scanline_last_fov_lat}  =  'finish' ;
    $globatt{geospatial_first_scanline_first_fov_lon} = 'start' ;
    $globatt{geospatial_first_scanline_last_fov_lon}  =  'finish' ;
    $globatt{geospatial_last_scanline_first_fov_lat}  = 'start' ;
    $globatt{geospatial_last_scanline_last_fov_lat}   =  'finish' ;
    $globatt{geospatial_last_scanline_first_fov_lon}  = 'start' ;
    $globatt{geospatial_last_scanline_last_fov_lon}   =  'finish' ;
    $globatt{geospatial_bounds} = 'polygon' ;
    $globatt{aggregate} = "$revision ($name)";
    foreach my $key (keys %globatt) {
      my $opcode = $globatt{$key};
      %globatt = aggregate_globatts($key, $opcode, \%headers, %globatt);
    }
    ### pretty_dump('%globatt', %globatt);
    # we can now write the header (but not the data, thus far uninterrogated)
    my $agg_cdl = join('.', $aggregated_filename, 'cdl');
    my $agg_nc4 = join('.', $aggregated_filename, 'nc');
    my $out_cdl = File::Spec -> catfile($destDir, $agg_cdl); 
    my $out_nc4 = File::Spec -> catfile($destDir, $agg_nc4); 
    # open the output file and write the "salutation"
    open (my $FH, ">$out_cdl") or die "$name\: Cannot open $out_cdl for writing\nAbort";
    print $FH "netcdf $aggregated_filename {\n";
    # construct and write the aggregated header
    my @header = aggregate_header(\%headers, \%newDim, \%varatt, %globatt);
    # REMOVE ME is a tag to replace unwanted coordinates on Pressure & Effective_Pressure
    foreach (@header) { ( /REMOVE ME/ ) and next; print $FH "$_\n"; }
    # write the new dimension data
    print $FH "data:\n\n";
    print $FH $newDim{dat}."\n";
    # outer loop over all remaining variables (in @varnames)
    foreach my $varname (@varnames) {
      my @data = ();
      if (grep {$_ eq $varname} qw/quality_information CrIS_Frequencies/) {
        @data = aggregate_data($varname, $files[0]);
      } else {
        @data = aggregate_data($varname, @files);
      }
      $data[0] = "\n$varname = ".$data[0];
      foreach (@data) { print $FH "$_\n"; }
    }
    # write the closing brace and close file
    print $FH "}\n";
    close($FH);
    # convert the .cdl file to netCDF4
    my $cmd = "$NCGEN -k4 -o $out_nc4 $out_cdl";
    (system($cmd) == 0) or warn "$name\: Could not run command $cmd\n";
    return;
  }

# -------------------------------
  sub aggregate_nucaps_retrievals
# ===============================
#  aggregate_nucaps_retrievals($sourceDir, $destDir, $revision);
  {
    # This is the top level routine for nucaps output netCDF4 file
    # aggregation that needs to be slotted into run_nucaps.pl
    my ($sourceDir, $destDir, $revision) = (shift, shift, shift);
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my @ccr_files = (sort(glob("$sourceDir/NUCAPS-CCR-AR*.nc")));
    my @edr_files = (sort(glob("$sourceDir/NUCAPS-EDR*.nc")));
    aggregate_nucaps_filegroup($destDir, $revision, @ccr_files);
    aggregate_nucaps_filegroup($destDir, $revision, @edr_files);
    return;
  }

# ---------------------
  sub aggregate_varatts
# =====================
# %varatt = aggregate_varatts($opcode, \%headers, %varatt);
  {
    # %varatt is a hash of replacements to make to variable attributes
    # $opcode is identifies the search strategy, basically.
    # %varatt is updated with a key=$attname and a value=full attribute line
    # Supported opcodes are 'quality', 'dimensions' and 'coordinates'.
    #
    # $opcode 'quality' triggers a very specific response, namely it searches
    # for and updates quality_information and its attributes that are counts
    # or statistics.
    #
    # $opcode 'coordinates' triggers a search for every occurance of 
    # 'coordinates = "Time Latitude Longitude"' and makes into 
    # 'coordinates = "Latitude Longitude" '
    # It also drops the attributes Pressure:coordinates and
    # Effective_Pressure:coordinates
    #
    # $opcode 'dimensions' changes 1D horizontal space dimension into
    # 2D AND makes vertical dimension 1D.
    my $opcode = shift;
    my %headers = %{(shift)};
    my (%varatt) = @_;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    ($opcode eq 'coordinates') and do {
      foreach my $key (sort keys %headers) {
        my @buffer = @{$headers{$key}};
        foreach (@buffer) {
          ( /\t\t/ ) or next;
          ( /coordinates = \"Time / ) and do {
            my $key = trim_string((split(/=/, $_))[0]);
            my $value = $_;
            $value =~ s/Time //g;
            $varatt{$key} = $value;
          };
        }
        $varatt{'Pressure:coordinates'} = 'REMOVE ME';
        $varatt{'Effective_Pressure:coordinates'} = 'REMOVE ME';
        last; # Only need to pass through one header for this task
      }
    };
    ($opcode eq 'dimensions') and do {
      foreach my $key (sort keys %headers) {
        my @buffer = @{$headers{$key}};
        foreach (@buffer) {
          ( /\t\t/ ) and next;
          ( /CrIS_Frequencies\(Number_of_CrIS_Frequencies\)/ ) and next;
          ( /\(/) or next;
          ( /\t/ ) and do {
            my $key = trim_string((split(/ /, $_))[1]);
            $key =~ s/\(/\\\(/g;  # Need to escape the brackets otherwise the
            $key =~ s/\)/\\\)/g;  #   search/replace fails in aggregate_header()
            my $value = $_;
            # exception processing to make vertical coordinates one dimensional
            $value =~ s/float Pressure\(Number_of_CrIS_FORs, /float Pressure\(/g;
            $value =~ s/float Effective_Pressure\(Number_of_CrIS_FORs, /float Effective_Pressure\(/g;
            # general processing to make 1D horizontal space into 2D space
            $value =~ s/Number_of_CrIS_FORs/Number_of_CrIS_scans, Number_of_CrIS_FOR/g;
            $varatt{$key} = $value;
          };
        }
        last; # Only need to pass through one header for this task
      }
    };
    ($opcode eq 'quality') and do {
      my ($Q1, $Q2, $Q3, $Q4) = ('N/A', 'N/A', 'N/A', 'N/A');
      my ($k1, $k2, $k3, $k4) = ('N/A', 'N/A', 'N/A', 'N/A');
      my (@q2, @q3, @q4) = ((), (), ());
      foreach my $key (sort keys %headers) {
        my @buffer = @{$headers{$key}};
        my $i = 0;
        foreach (@buffer) {
          $i++; ( /char quality_information/ ) and last;
        }
        $Q1 = $buffer[$i];
        $k1 = trim_string((split(/=/, $Q1))[0]);
        $Q2 = $buffer[$i+1];
        $k2 = trim_string((split(/=/, $Q2))[0]);
        my $num = trim_string((split(';', (split(/=/, $Q2))[1]))[0]);
        is_integer($num) or 
          die "$name\: total_number_retrievals non-integer\nAbort";
        push(@q2, $num);
        $Q3 = $buffer[$i+2];
        $k3 = trim_string((split(/=/, $Q3))[0]);
        my $opt = trim_string((split('f', (split(/=/, $Q3))[1]))[0]);
        looks_like_number($opt) or 
          die "$name\: percentage_optimal_retrievals not a number\nAbort";
        push(@q3, $opt);
        $Q4 = $buffer[$i+3];
        $k4 = trim_string((split(/=/, $Q4))[0]);
        my $bad = trim_string((split('f', (split(/=/, $Q4))[1]))[0]);
        looks_like_number($bad) or 
          die "$name\: percentage_optimal_retrievals not a number\nAbort";
        push(@q4, $bad);
      }
      my $NUM = sum(@q2);
      my $OPT = sprintf(" %.2ff ;", 0);
      my $BAD = sprintf(" %.2ff ;", 0);
      ($NUM > 0) and do {
        my ($WOPT, $WBAD) = (0, 0);
        for (my $j=0; $j<=$#q2; $j++) {
          $WOPT += $q2[$j] * $q3[$j];
          $WBAD += $q2[$j] * $q4[$j];
        }
        $OPT = sprintf(" %.2ff ;", $WOPT/$NUM);
        $BAD = sprintf(" %.2ff ;", $WBAD/$NUM);
       };
      $Q1 =~ s/granule/aggregate/g;
      $Q2 = join ('=', (split(/=/, $Q2))[0], ' '.$NUM.' ;');
      $Q3 = join ('=', (split(/=/, $Q3))[0], $OPT);
      $Q4 = join ('=', (split(/=/, $Q4))[0], $BAD);
      $varatt{$k1} = $Q1;
      $varatt{$k2} = $Q2;
      $varatt{$k3} = $Q3;
      $varatt{$k4} = $Q4;
    };
    return %varatt;
  }

# ----------------------
  sub aggregate_varnames
# ======================
# @varnames = aggregate_varnames(%headers);
  {
    # @varnames is a list of variable names taken from just the
    # first file whose header has been read into %headers
    my (%headers) = @_;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my @varnames = ();
    foreach my $key (sort keys %headers) {
      my @buffer = @{$headers{$key}};
      foreach (@buffer) {
        ( /global attributes:/ ) and last;
        ( /\=/ ) and next;    # rejects dimensions
        ( /\t\t/ ) and next;  # rejects attributes
        ( /\t/ ) and do {     # they start with one tab
          my $value = (split(/ /, $_))[1];
          $value  = (split(/\(/, $value))[0];
          ($value eq 'Scan_Line') and next;
          ($value eq 'CrIS_FORs') and next;
          push(@varnames, $value);
        };
      }
      last; # Only need to pass through one header for this task
    }
    return @varnames;
  }

# --------------------
  sub compute_kai_skip
# ====================
# $kai_skip = compute_kai_skip($options{mdrstart}, $iasidbfile);
  {
    # $kai_skip is the number of MDR records to skip to achieve the
    # selected $options{mdrstart} for file $iasidbfile.
    my ($mdrstart, $file) = (shift, shift);
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my $skip = 0;
    if (is_integer($mdrstart)) {
      $skip = $mdrstart;
    } else {
      (grep {$_ eq $mdrstart} qw/E M/) and do {
        my $cmd = "$KAI --out TMP.iasi --in $file 2>/dev/null";
        my $dribble = `$cmd`;
        unlink('TMP.iasi') or die "$name\: cannot delete TMP.iasi\nAbort";
        foreach (split("\n",$dribble)) {
          (index($_, '    MDR   :') < 0) or do {
            my $mdr = trim_string((split(':', $_))[1]);
            (is_integer($mdr)) and do {
              my $n = 0;
              my @seq = (23,22) x 99;
              while (($mdr - $seq[$n])> 0) {
                $mdr -= $seq[$n++];
              }
              ($mdrstart eq 'E') and $skip = $mdr;
              ($mdrstart eq 'M') and $skip = int($mdr/2);
            };
          };
        }
      };
    }
    return $skip;
  }

# -------------------
  sub get_commandline
# ===================
# %options = get_commandline($revision, $pwd);
  {
    # If things go bad in here, jump to usage() which terminates with exit(9).
    # Note that I had to hack in the Python-like -vv & -vvv support.
    # This routine assumes standard module Getopt::Long has been loaded.
    # abs_path() might be dangerous for the purpose of setting the work (and,
    # equivalently, the output) directory - see this post:
    # http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=257568
    my ($rev, $pwd) = (shift, shift);
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my $codeset = langinfo(CODESET);
    @ARGV = map{decode $codeset, $_} @ARGV;
    foreach (@ARGV) { s/\?/-\?/g; }
    my %options = ();
    GetOptions(\%options,
      'version|V',
      'help|?',
      'nothing|n',
      'input_files|i=s',
      'sat|s=s',
      'mdrstart|m=s',
      'work|w=s',
      'dynanc|d=s',
      'reproc|r=i',
      'processors|p=i',
      'verbose|v+', 'vv', 'vvv',
      'xfiles|x',
      'aggregate|a!',
      'quarantine|q!',
      'clean|c!') or usage();
    defined($options{version}) and die "$rev\n";
    defined($options{help}) and usage();
    $options{nothing} = defined($options{nothing}) ? $options{nothing} : 0;
    $options{xfiles} = defined($options{xfiles}) ? $options{xfiles} : 0;
    defined($options{input_files}) or usage();
    defined($options{sat}) or usage();
    (grep {$_ eq $options{sat}} qw/metopA metopB/) or do {
      print "\nERROR --- $name\: Unsupported MISSION\n"; usage(); };
    $options{mdrstart} = defined($options{mdrstart}) ? $options{mdrstart} : 'M';
    (grep {$_ eq $options{mdrstart}} qw/S M E/) or do {
      (is_integer($options{mdrstart})) or do {
        print "\nERROR --- $name\: Unsupported mdrstart\n"; usage(); }; };
    (defined($options{work})) or $options{work} = $pwd;
    (defined($options{reproc})) or $options{reproc} = 0;
    (defined($options{dynanc})) or 
      $options{dynanc} = File::Spec -> catfile($ENV{CSPP_NUCAPS_IASI_HOME}, 
        'data', 'dynamic_ancillary');
    foreach (qw/dynanc work/) {
      $options{$_} = File::Spec -> catfile(abs_path(dirname($options{$_})),
        basename($options{$_}))
    }
    (defined($options{processors})) or $options{processors} = 1;
    ($options{processors} < 1) and usage();
    $options{verbose} = defined($options{verbose}) ? $options{verbose} : 0;
    $options{vv} = defined($options{vv}) ? $options{vv} : 0;
    $options{vvv} = defined($options{vvv}) ? $options{vvv} : 0;
    $options{verbose} = $options{verbose} + 2 * $options{vv};
    $options{verbose} = $options{verbose} + 3 * $options{vvv};
    $options{verbose} = min($options{verbose}, 3);
    delete @options{'vv', 'vvv'};
    $options{clean} = defined($options{clean}) ? $options{clean} : 1;
    $options{aggregate} = defined($options{aggregate}) ? $options{aggregate} : 0;
    $options{quarantine} = defined($options{quarantine}) ? $options{quarantine} : 0;
    ($options{reproc}) and $options{quarantine} = 1; # force no LFTP fetch.
    return %options;
  }

# -----------------------------------------------------------
  sub is_integer { defined $_[0] and $_[0] =~ /^[+-]?\d+$/; }
# ===========================================================

# ---------------
  sub lftp_mirror
# ===============
# $status = lftp_mirror($site, $file, $local, %options);
  {
    # executes lftp in mirror mode, but if NOT $options{quarantine}
    my ($site, $file, $local) = (shift, shift, shift);  
    my (%options) = @_;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my $v = '';
    ($options{verbose} == 3) and $v = '-v';
    -e "$local/$file" and do {
      ($options{verbose} == 3) and print "$name\: $file FOUND in local repo\n";
      return 1;
    };
    ($options{quarantine}) and do {
      ($options{verbose} >= 1) and 
        warn "$name\: WARNING: $file NOT FOUND in local repo AND --quarantine";
      return 0;
    };
    open(my $fh, '>script_file.lftp') or die "$name\: Cannot open script_file.lftp $!";
    print $fh "set dns:order inet inet6\n";            # default is 'inet6 inet'
    #print $fh "set ssl:verify-certificate no\n";       # default is ‘yes'
    #print $fh "set net:timeout 1\n";                   # default is '5'
    #print $fh "set net:max-retries 5\n";               # default is '10'
    #print $fh "set net:persist-retries 1\n";           # default is '5'
    #print $fh "set net:reconnect-interval-base 6\n";   # default is '60'
    print $fh "mirror $v -L -n --no-empty-dirs $site -I $file $local\n";
    close($fh);
    my $cmd = "flock $local $LFTP -f script_file.lftp";
    system($cmd) == 0 or do {
      ($options{verbose} >= 1) and warn "$name\: WARNING: $cmd failed";
      return 0; };
    return 1;
  }

# -------------------
  sub make_CPU_STATUS
# ===================
# make_CPU_STATUS(%options)
  {
    # Made into a function to make easier its recreation between DAP steps.
    my (%options) = @_;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    system("echo CPU_01 Idle > CPU_STATUS") == 0 or 
      die "$name\: Could not create CPU_STATUS file\nAbort";
    foreach (2..$options{processors}) {
      my $cpu = sprintf("CPU_%02d Idle", $_);
      my $cmd = "echo $cpu >> CPU_STATUS";
      system($cmd) == 0 or die "$name\: Could not create CPU_STATUS file\nAbort";
    }
  }

# ----------------------
  sub make_L1B_bigEndian
# ======================
# $outfile = make_L1B_bigEndian($instrument, $infile);
  {
    # $instrument is one of 'amsua' or 'mhs'
    # $infile is the name of the L1B file to test/convert endian
    # $outfile is the big endian output filename created
    #
    # $outfile is extracted from the header of $infile, bytes 23-64 inclusive
    # (42 bytes). The reason for creating $outfile in WORK_DIR/tempdir is so
    # that we don't clobber data files in WORK_DIR if the user happens to
    # specify WORK_DIR as the place where the original L1B files are stored.
    # For files named like "amsual1b_noaa18_20131014_2145_43295.l1b" no problem,
    # but for files named like "NSS.AMAX.NN.D06032.S0122.E0317.B0361819.WI",
    # problem since this is the same as the filename stored in its header!
    #
    # NOTE: We coalesce repeated character variables to strings in the $Lhdr and
    # $Lrec buffer specification (because we are not actually applying any logic
    # to characters), but only after 'a42a8' in $Lhdr, since the a42 field is to
    # be used to define $outfile.
    #
    my ($instrument, $infile) = (shift, shift);
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my ($Lhdr, $Bhdr, $Lrec, $Brec, $count, $buf) = undef;
    my ($reclen, $year, $iyear, $doy, $idoy, $iout) = undef;
    ($instrument eq 'amsua') and do {
      ($reclen, $iout, $iyear, $idoy) = (2560, 11, 22, 23);
      $Lhdr = 'a3av9a42a8v4Vv2V2v2Vv3v5Vv2V2v2Vv14Va16V4v106V115a8v8V363v6Vv2V46';
      $Lrec = 'v4Vv2V4v16V95v6V2v94V61v620Vv2V2a28Vv156Vv2V2a32v6V'; };
    ($instrument eq 'mhs') and do {
      ($reclen, $iout, $iyear, $idoy) = (3072, 11, 22, 23);
      $Lhdr = 'a3av9a42a8v4Vv2V2v2Vv8Vv2Vv15a16v51V53a8v8V92v436Vv2V6v642';
      $Lrec = 'v4Vv2VvaaV2v14V30v8V2v274V180v544V2v48V2a78v9V5a52v12Vv102'; };
    defined($reclen) or die "$name\: Instrument $instrument not supported\nAbort";
    $Bhdr = $Lhdr; $Bhdr =~ s/v/n/g; $Bhdr =~ s/V/N/g;
    $Brec = $Lrec; $Brec =~ s/v/n/g; $Brec =~ s/V/N/g;
    open my $fh, "<$infile"; binmode $fh;
    $count = read($fh, $buf, $reclen);
    ($count == $reclen) or die "$name\: Header count != reclen $infile\nAbort";
    my @data = unpack($Lhdr, $buf);
    ($year, $doy) = ($data[$iyear], $data[$idoy]);
    my $Lendian = ($year > 1960 and $year < 2099 and $doy > 0 and $doy < 367);
    @data = unpack($Bhdr, $buf);
    ($year, $doy) = ($data[$iyear], $data[$idoy]);
    my $Bendian = ($year > 1960 and $year < 2099 and $doy > 0 and $doy < 367);
    ($Lendian xor $Bendian) or die "$name\: $infile endian indeterminate\nAbort";
    my $outfile = $data[$iout];
    $outfile =~ s/^\s+//; $outfile =~ s/\s+$//; $outfile =~ s/ //g;
    (length($outfile) != 42) and die "$name\: L1B filename $outfile funky\nAbort";
    open my $FH, ">$outfile"; binmode $FH;
    ($Lendian) and $buf = pack($Bhdr, unpack($Lhdr, $buf));
    print $FH $buf;
    while (!eof($fh)) {
      $count = read($fh, $buf, $reclen);
      ($count == $reclen) or die "$name\: Record count != reclen $infile\nAbort";
      ($Lendian) and $buf = pack($Brec, unpack($Lrec, $buf));
      print $FH $buf;
    }
    close $fh; close $FH;
    return $outfile;
  }

# --------------------
  sub make_ISRP_CONFIG
# ====================
# make_ISRP_CONFIG(%options);
  {
    # Creates the ISRP_CONFIG file in $tmpdir.
    # This configuration can be used for multiple days.  I am going to create
    # from scratch, i.e. not read and modify a template.  This is minimalist -
    # comment lines are gone, if the system runs without a certain entry then
    # I omit it.  If you want to see an annotated example, go check out the
    # original $CSPP_NUCAPS_IASI_HOME/$CSPP_NUCAPS_IASI_REVNAME/ISRP_CONFIG
    # N.B. $PLATFORM_SYS and sys_files have been symlinked in the temporary
    # working directory $tmpdir because long filepaths to support files causes
    # some (at least one) DAP executable to overflow its command line buffer.
    my (%options) = @_;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my ($PLATFORM_NAME, $PLATFORM_ID, $PLATFORM_SYS) = (undef, undef, undef);
    ($options{sat} eq 'metopA') and do {
      $PLATFORM_NAME = 'MetOpA';
      $PLATFORM_ID='M02';
      $PLATFORM_SYS='metopA_sys_files'; };
    ($options{sat} eq 'metopB') and do {
      $PLATFORM_NAME = 'MetOpB';
      $PLATFORM_ID='M01';
      $PLATFORM_SYS='metopB_sys_files'; };
    open(my $fh, ">ISRP_CONFIG") or die "$name\: Cannot open ISRP_CONFIG for write.\nAbort";
    print $fh "USE_AVHRR=OFF\n";
    print $fh "GET_FILES_FROM_ARCHIVES=OFF\n";
    print $fh "CLEAN_UP_FLAG=OFF\n";
    print $fh "CREATE_OLR_FLAG=OFF\n";
    print $fh "CREATE_PC_FLAG=OFF\n";
    print $fh "DATE_CHECK=ON\n";
    print $fh "OPS_MACHINE=localhost\n";
    print $fh "OPS_MAIL=nobody\@localhost\n";
    print $fh "OPS_LOGS=$tmpdir/logs\n";
    print $fh "OPSDATA=$tmpdir/retrieval\n";
    print $fh "PERL_OPS=$DAP/scripts/perl\n";
    print $fh "KSH_OPS=$DAP/scripts/ksh\n";
    print $fh "OPSBIN=$DAP/bin\n";
    print $fh "L1C_DATA=$tmpdir/input/L1C\n";
    print $fh "L1B_DATA=$tmpdir/input/L1B\n";
    print $fh "MODEL_DATA=$tmpdir/input/GFS\n";
    print $fh "TIME_DIR=$tmpdir/time_files\n";
    print $fh "CPU_STATUS_FILE=$tmpdir/CPU_STATUS\n";
    print $fh "EIGEN_DIR=$tmpdir/sys_files/eigen\n";
    print $fh "EIGEN_DATE=070804\n";
    print $fh "L2_CONFIG=$tmpdir/$PLATFORM_SYS/retrieval\n";
    print $fh "L1C_METADATA_TEMPLATE=$tmpdir/$PLATFORM_SYS/metadata/IASI_Granule_Metadata_Template.xml\n";
    print $fh "OPSCDL=$tmpdir/sys_files/cdl\n";
    print $fh "BUFR_TABS=$tmpdir/sys_files/bufr\n";
    print $fh "DEM_FILE=$tmpdir/sys_files/dem/Global_DEM.bin\n";
    print $fh "DEM_LAT_PTS=21600\n";
    print $fh "DEM_LON_PTS=43200\n";
    print $fh "LAC_DIR=$tmpdir/sys_files/lac\n";
    print $fh "LAC_DATE=090612\n";
    print $fh "LAC_FLAG=LAC_ON\n";
    print $fh "OLR_DIR=$tmpdir/sys_files/olr\n";
    print $fh "OLR_COR=$tmpdir/sys_files/olr/rad_corr_reg_coef_20boxcar_airsv10ab.asc\n";
    print $fh "OLR_REG=$tmpdir/sys_files/olr/olr_reg_coef_cv005_20boxcar.asc\n";
    print $fh "OLR_METADATA_TEMPLATE=$tmpdir/$PLATFORM_SYS/metadata/IASI_OLR_Granule_Metadata_Template.xml\n";
    print $fh "AVH_WTS=$tmpdir/sys_files/avhrr/iasiwgt.dat\n";
    print $fh "PREPROC_TEMPLATE=$tmpdir/$PLATFORM_SYS/retrieval/surface/preprocessor.cfg\n";
    print $fh "CCR_METADATA_TEMPLATE=$tmpdir/$PLATFORM_SYS/metadata/IASI_CCR_Granule_Metadata_Template.xml\n";
    print $fh "L2_METADATA_TEMPLATE=$tmpdir/$PLATFORM_SYS/metadata/IASI_L2_Granule_Metadata_Template.xml\n";
    print $fh "NCBIN=".dirname($NCDUMP)."\n";
    print $fh "UUIDBIN=$UUIDGEN\n";
    # This location speficied for wgrib2 is not honored, needs to be in OPSBIN=$DAP/bin, remove from config.
    # print $fh "WGRIB2_LOC=".dirname($WGRIB2)."\n";
    print $fh "PLATFORM_NAME=$PLATFORM_NAME\n";
    print $fh "PLATFORM_ID=$PLATFORM_ID\n";
    print $fh "DIST_FLAG_PFS=OFF\n";
    print $fh "DDS_DIST_DIR=none\n";
    print $fh "BYTESWAP_FLAG=YES\n";
    close($fh) or die "$name\: Cannot close ISRP_CONFIG.\nAbort";
    return;
  }

# --------------------
  sub nucaps_avn_files
# ====================
# %avn = nucaps_avn_files($gmtbeg, %dynanc_loc);
  {
    # $gmtbeg is the beginning time of the CrIS SDR
    # %dynanc_loc is hash of dynamic ancillary data locations, local & remote
    # %avn is a hash of URL, LOCAL repo directory plus AVN file pairs, keyed
    # 'table' and 'spare', where 'table' is original AVN truth table as defined
    # by Chris Barnet, and 'spare' is the fallback possition if 'table' options
    # are unavailable. This fact cannot be known now, only when we go to the
    # ancillary data server or, if --quarantine is invoked, when we check the
    # local repo. Because the date is in both the URL and the LOCAL directory
    # name, and this could be different in the fallback positions, we have to
    # build all of $url, $local, $avn1 and $avn2 and return them in the hash;
    # I am going to just put then in 4 element vector in that order. The
    # bounding gfs files to use are for NUCAPS processing of CrIS SDR with
    # input $gmtbeg.
    my $gmtbeg = shift;
    my (%dynanc_loc) = @_;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    # AVN truth table as defined by Chris Barnet
    my %avn_truth_table = ('00z' => [('18z', 'F06', 'F09', 'yesterday')],
                           '03z' => [('00z', 'F03', 'F06', 'today')],
                           '06z' => [('00z', 'F06', 'F09', 'today')],
                           '09z' => [('06z', 'F03', 'F06', 'today')],
                           '12z' => [('06z', 'F06', 'F09', 'today')],
                           '15z' => [('12z', 'F03', 'F06', 'today')],
                           '18z' => [('12z', 'F06', 'F09', 'today')],
                           '21z' => [('18z', 'F03', 'F06', 'today')]);
    # AVN truth table fallback position 
    my %avn_truth_spare = ('00z' => [('12z', 'F12', 'F15', 'yesterday')],
                           '03z' => [('18z', 'F09', 'F12', 'yesterday')],
                           '06z' => [('18z', 'F12', 'F15', 'yesterday')],
                           '09z' => [('00z', 'F09', 'F12', 'today')],
                           '12z' => [('00z', 'F12', 'F15', 'today')],
                           '15z' => [('06z', 'F09', 'F12', 'today')],
                           '18z' => [('06z', 'F12', 'F15', 'today')],
                           '21z' => [('12z', 'F09', 'F12', 'today')]);
    gmtime($gmtbeg);
    my $YYYY = sprintf("%04d", $tm_year + 1900);
    my $ddd = sprintf("%03d", $tm_yday + 1);
    my $hh = sprintf("%02d", $tm_hour);
    my $key = sprintf("%02dz" , $hh - $hh % 3);
    my %avn = ();
    foreach (qw/table spare/) {
      my ($anal, $fct1, $fct2, $relday) = @{$avn_truth_table{$key}};
      /spare/ and ($anal, $fct1, $fct2, $relday) = @{$avn_truth_spare{$key}};
      my $avn1 = $dynanc_loc{GFSTEMPLATE};
      my $avn2 = $dynanc_loc{GFSTEMPLATE};
      $hh = substr($anal, 0, 2);
      my ($ff1, $ff2) = (substr($fct1, 1, 2), substr($fct2, 1, 2));
      $avn1 =~ s/%f/$ff1/g;
      $avn2 =~ s/%f/$ff2/g;
      ($relday eq 'yesterday') and $ddd--;
      gmtime(timegm_nocheck(0, 0, $hh, $ddd, 0, $YYYY));
      $YYYY = sprintf("%04d", $tm_year + 1900);
      my $MM = sprintf("%02d", $tm_mon + 1);
      my $DD = sprintf("%02d", $tm_mday);
      $ddd = sprintf("%03d", $tm_yday + 1);
      $hh = sprintf("%02d", $tm_hour);
      my $YY = substr($YYYY, 2 , 2);
      my @tokens = qw/ %Y   %y   %m    %j   %d   %H/;
      my @values = ($YYYY, $YY, $MM, $ddd, $DD, $hh);
      my $site = $dynanc_loc{'GFSSITE'};
      my $proto = $dynanc_loc{'GFSPROTO'};
      my $local = $dynanc_loc{'GFSLOCAL'};
      my $remote = $dynanc_loc{'GFSREMOTE'};
      for (my $j=0; $j<=$#tokens; $j++) {
        $local =~ s/$tokens[$j]/$values[$j]/g;
        $remote =~ s/$tokens[$j]/$values[$j]/g; 
        $avn1 =~ s/$tokens[$j]/$values[$j]/g; 
        $avn2 =~ s/$tokens[$j]/$values[$j]/g; 
      }
      # ********** MAKE PATH SO THAT WARNINGS CAN BE REPORTED **********
      if (! -d $local) {
        eval { mkpath($local) };
        if ($@) { warn "$name\: WARNING $local cannot be made"; }
      }
      my $url = "$proto$site/$remote";
      @{$avn{$_}} = ($url, $local, $avn1, $avn2);
    }
    return (%avn);
  }

# -----------------------
  sub populate_dynanc_gfs
# =======================
# populate_dynanc_gfs(\%dynanc_loc, \%options, @gmt);
  {
    # %dynanc_loc is remote/local information for GFS dynamic ancillary data.
    # %options are the command line options.
    # @gmt is the IASI granule start times.
    # N.B. I am goin to 'warn' in here where I woould normally 'die' so that
    # if multiple days are being processed, there rmeains a chance the next
    # days will process. (But I 'die' if input/GFS cannot be made).
    # During the execution of this function, the $PWD is assumed to be $tmpdir.
    # Function nucaps_avn_files() written for CSPP_NUCAPS is very targeted,
    # identifying only the GFS files that bound the granules.  ISRP seems to
    # want the whole day of GFS, so I will just fetch it all, but not bail if
    # some are not yet ready.  I will still use nucaps_avn_files() to seed
    # my solution, since most of the logic is in there.        JED 10-May-2018
    my %dynanc_loc = %{(shift)};
    my %options = %{(shift)};
    my @gmt = @_;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    # Our server version of GFS 1-deg file is date/time encoded and unique
    my %gfshash = ();
    foreach (@gmt) {
      my %avn = nucaps_avn_files($_, %dynanc_loc);
      foreach (qw/table spare/) {
        my ($url, $local, $avn1, $avn2) = @{$avn{$_}};
        foreach (($avn1, $avn2)) {
          @{$gfshash{$_}} = ($url, $local);
        }
      }
    }
    # %date is keyed by url which will include a unique date
    my %date = ();
    foreach my $avn (sort keys %gfshash) {
      my ($url, $local) = @{$gfshash{$avn}};
      @{$date{$url}} = ($local, $avn);
    }
    ($options{verbose} == 3) and do {
      print "--------------\n";
      print "GFS %date keys\n";
      print "==============\n";
      foreach (sort keys %date) { print "$_\n"; }
      print "GFS %date keys _END_\n"; };
    # fetch filenames 
    my $template = "$dynanc_loc{'GFSREMOTE'}$dynanc_loc{'GFSLOCAL'}$dynanc_loc{'GFSTEMPLATE'}";
    my $preamble = "$dynanc_loc{'GFSPROTO'}$dynanc_loc{'GFSSITE'}";
    my %tokens = ('%Y'=>4, '%y'=>2, '%m'=>2, '%j'=>3, '%d'=>2,
                  '%H'=>2, '%M'=>2, '%S'=>2, '%F'=>3, '%f'=>2);
    foreach my $url (sort keys %date) {
      my ($local, $avn) = @{$date{$url}};
      my $n = length($preamble);
      my $proto = $dynanc_loc{'GFSTEMPLATE'};
      my $match = substr($url.$local.$avn, $n+1);
      foreach my $key (keys %tokens) {
        (grep {$_ eq $key} qw/%H %f %F/) and next;
        my $fmt = '%'.sprintf("0%sd", $tokens{$key});
        my $found = token_value($key, $match, $template);
        ($found < 0) and next;
        my $value = sprintf($fmt, $found);
        $proto =~ s/$key/$value/g;
      }
      # These are in https://jpssdb.ssec.wisc.edu/cspp_v_2_0/ancillary/ structure
      foreach my $hh ((0,6,12,18)) {
        my $fmt = '%'.sprintf("0%sd", $tokens{'%H'});
        my $value = sprintf($fmt, $hh);
        my $base = $proto;
        $base =~ s/%H/$value/g;
        foreach my $ff ((3,6,9,12)) {
          my $avn = $base;
          foreach (qw/%F %f/) {
            $fmt = '%'.sprintf("0%sd", $tokens{$_});
            $value = sprintf($fmt, $ff);
            $avn =~ s/$_/$value/g;
          }
          # Get GFS files and make links just like I do for CSPP_NUCAPS
          my $waitawhile = $SLEEP_GFS_MINUTES;
          my $target = File::Spec -> catfile($local, $avn);
          -e $avn or do {
            lftp_mirror($url, $avn, $local, %options);
            -e $target and symlink($target, $avn) or
              warn "$name\: WARNING: Cannot link to $target";
            ($options{quarantine}) and next; # don't need multiple attempts at local repo
            -e $avn and next; # we have $target and link, so we are done for this $avn
            my $j = $waitawhile; # so $j is still in scope after the following loop
            for ($j=$waitawhile; $j>0; $j--) {
              ($options{verbose} >= 1) and warn "$name\: SLEEPING at minute $j waiting for $avn\n";
              sleep 60;
              lftp_mirror($url, $avn, $local, %options);
              -e $target or next;
              symlink($target, $avn) or warn "$name\: WARNING: Cannot link to $target";
              last;
            }
            $waitawhile = $j;
          };
        }
      }
    }
    # Move the links into the directory structure & filenames expected by ISRP
    my $aviation = File::Spec -> catfile('input', 'GFS', 'AVIATION');
    mkpath($aviation) or die "$name\:  WORK_DIR/$aviation cannot be made.\nAbort";
    foreach (sort(glob("gfs.*"))) {
      my ($YYYY, $MM, $DD) = ('20'.substr($_, 8, 2), substr($_, 10, 2), substr($_, 12, 2));
      my $newdir = File::Spec -> catfile($aviation, $YYYY, $MM, $DD);
      -e $newdir or mkpath($newdir) or warn "$name\: WORK_DIR/$newdir cannot be made.\nAbort";
      my ($hh, $ff) = (substr($_, 5, 2), substr($_, 20, 2));
      my $target = 'gfs.t'.$hh.'z.pgrb2f'.$ff;
      move($_, "$newdir/$target") or warn "$name\: Could not move $_ to $newdir/$target\n";
    }
    return;
  }

# ---------------
  sub pretty_dump
# ===============
# pretty_dump($title, %hash);
  {
    # Do a "pretty print" of %hash by effectively grepping Dumper().
    # Assumes standard module Data::Dumper has been loaded.
    my $title = 'DEBUG '.shift;
    my (%hash) = @_;
    print '-' x length($title)."\n";
    print "$title\n";
    print '=' x length($title)."\n";
    foreach (split('\n', Dumper(\%hash))) {
      chomp;
      ((/VAR1/) or (/\}\;/)) and next;
      $_ =~ s/  //g; # shorten up the whitespace
      print "$_\n";
    }
    print "$title _END_\n";
  }

# ---------------
  sub token_value
# ===============
# $value = token_value($token, $file, $mask);  ...example follows...
# $value = token_value('%j', 't1.11332.1500.mod04.hdf', 't1.%y%j.%H%M.mod04.hdf');
  {
    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    #                                                                           #
    #   WARNING: This function not used in NUCAPS but kept as token reference   #
    #                                                                           #
    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    #
    # Takes the filename $file (within which tokens have already been expanded),
    # and the $mask (within which tokens are unexpanded) and, for a given $token
    # return the $value extracted from $file. Here is an example ...
    #   $value = token_value('%j', 't1.11332.1500.mod04.hdf', 't1.%y%j.%H%M.mod04.hdf');
    # which would return $value as '332'.
    # These are the supported tokens, returns the first occurance if more than one:
    #   %Y  Year with century. (1582-9999)
    #   %y  Year without century (00–99).
    #   %m  Month as decimal number (01–12).
    #   %j  Day of year as decimal number (001–366).
    #   %d  Day of the month as decimal number (01–31).
    #   %H  Hours as decimal number (00–23).
    #   %M  Minute as decimal number (00–59).
    #   %S  Second as decimal number (00–59).
    #   %F  3-digit forecast hour (000-999).
    #   %f  2-digit forecast hour (00-99).
    my $token = shift;
    my $file = shift;
    my $mask = shift;
    my $value = -1;
    my $expand = $mask;
    my @tokens = qw/  %Y %y %m  %j %d %H %S %M  %F %f/;
    my @widths =   (   4, 2, 2,  3, 2, 2, 2, 2,  3, 2);
    my $width = 0;
    for (my $j=0; $j<=$#tokens; $j++) {
      ($token eq $tokens[$j]) and do { $width = $widths[$j]; next; };
      my $blank = 'X' x $widths[$j];
      $expand =~ s/$tokens[$j]/$blank/g; 
    }
    my $i = index($expand, $token);
    if ($i > -1) {
      $value = substr($file, $i, $width);
    }
    return $value;
  }

# ---------------
  sub trim_string
# ===============
# $new_string = trim_string($old_string);
  {
    my $string = shift;
    $string =~ s/^\s+//;
    $string =~ s/\s+$//;
    return $string;
  }

# ---------
  sub usage
# =========
# usage();
  {
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    open(my $fh, "<$0");
    while (<$fh>) { chomp;
      if ( /START-HELP/ ) {
        while (<$fh>) { chomp;
          ( /END-HELP/ ) and do { close($fh); exit(9); };
          $_ =~ s/\$CSPP_NUCAPS_IASI_HOME/$ENV{CSPP_NUCAPS_IASI_HOME}/g;
          my $help = (split(/\#/, $_))[1]; # remove leading hash
          $help and print "$help\n" or print "\n";
        }
      }
    }
    die "$name\: ERROR! Have you tinkered with help in script header?\nAbort";
  }

# --------------------
  sub validate_options
# ====================
# %options = validate_options($filemask, %options);
  {
    # $filemask is for globbing the nominated satellite data.
    # %options are the command line options to "validate".
    # Fatal errors in here terminate with die without showing usage() message.
    # The most important task is to convert $options{input_files} to a scalar
    # $options{input_dir} and filename list put back in $options{input_files}.
    # This is also where the working and ancillary data directories are made
    # if they do not exist, and their writable status verified.
    my $filemask = shift;
    my (%options) = @_;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my $globber = $options{input_files};
    my $altglob = $globber;
    (-d $options{input_files}) and do {
      my $base = $options{input_files};
      $globber = File::Spec -> catfile($base, $filemask); };
    my @input_files = sort(glob($globber));
    ($#input_files < 0) and die "$name\: INPUTFILES not found.\nAbort";
    $options{input_dir} = dirname(abs_path($input_files[0]));
    $options{input_files} = ();
    foreach (@input_files) { push(@{$options{input_files}}, basename($_)); }
    ($options{reproc} > 2) and do {
      die "$name\: --reproc=GRIB $options{reproc} is not valid.\nAbort"; };
    (-d $options{work}) or do {
      ($options{verbose} >=2) and print "Making WORK_DIR $options{work}\n";
      mkpath($options{work}) or
        die "$name\: WORK_DIR $options{work} cannot be made.\nAbort"; };
    (-w $options{work}) or
      die "$name\: WORK_DIR $options{work} not writable.\nAbort";
    ($options{quarantine}) or do {
      (-d $options{dynanc}) or do {
        ($options{verbose} >=2) and print "Making dynanc DIR $options{dynanc}\n";
        mkpath($options{dynanc}) or
          die "$name\: DYNANC dir $options{dynanc} cannot be made.\nAbort"; };
      (-w $options{dynanc}) or
        die "$name\: DYNANC dir $options{dynanc} not writable.\nAbort"; };
    return %options;
  }

